"""
This is a package for using sets of orthogonal functions/polynomials. Currently, it includes Chebyshev, Legendre, and Gegenbauer polynomials. It also has real, two-dimensional spherical harmonics.

Installing/Using
----------------

To install the package, you can

.. code-block::

    > pip install orthopoly

or you can download/clone the repository, put the top directory in your `sys.path`, then import it.


chebyshev
---------

The :mod:`~orthopoly.chebyshev` module is pretty well developed. It includes many functions for evaluating the Chebyshev polynomials and their derivatives. The module also includes higher level functions for setting up the elements needed to solve boundary value problems (using the pseudospectral method), interpolate using a Chebyshev grid, and perform the spatial discretization of a PDE solver. Generally, these methods are very well suited to smooth problems.

For performing a discrete Chebyshev transform (generating a Chebyshev expansion from a set of points in 1D), the :func:`~orthopoly.chebyshev.cheby_coef_setup` function can be used. It allows one of the boundary conditions to be the value of the expansion's first derivative and returns a matrix allowing expansion coefficents to be computed by solving a linear system. When there are no derivatives, the transform can be computed (on the appropriate :func:`grid points <orthopoly.chebyshev.cheby_grid>`) with a discrete cosine transform (DCT). The :func:`~orthopoly.chebyshev.cheby_dct` function does this and :func:`~orthopoly.chebyshev.cheby_dct_setup` sets up the grid as well.

Information about the Chebyshev polynomials is widely available, but a few particularly helpful references are below. The Boyd book is especially good.

* Boyd, John P. *Chebyshev and Fourier spectral methods*. Courier Corporation, 2001.
* Fornberg, Bengt. *A practical guide to pseudospectral methods. Vol. 1*. Cambridge university press, 1998.
* Canuto, Claudio, et al. *Spectral methods*. Springer-Verlag, Berlin, 2006.

gegenbauer
----------

Gegenbauer polynomials :math:`C_n^m(x)` are generalizations of Chebyshev and Legendre polynomials. However, chebyshev polynomials of the first kind are implemented by other methods in the :mod:`~orthopoly.chebyshev` module, and cannot be computed by the functions in this module.

legendre
--------

This module implements the Associated Legendre Polynomials, :math:`P_n^m(x)`, and their first two derivatives in support of the :mod:`~orthopoly.spherical_harmonic` module. If :math:`m=0`, they reduce to the unassociated Legendre polynomials.

spherical_harmonic
------------------

The :mod:`~orthopoly.spherical_harmonic` module provides functions for evaluating the real, two-dimensional (surface), orthonormal, spherical harmonics. From the associated Legendre polynomials, the :func:`spherical harmonics <orthopoly.spherical_harmonic.sph_har>`, their :func:`gradients <orthopoly.spherical_harmonic.grad_sph_har>`, and their :func:`Laplacians <orthopoly.spherical_harmonic.lap_sph_har>` can be evaluated. The module also contains some functions for creating grids on the sphere (:func:`regular <orthopoly.spherical_harmonic.grid_regular>`, :func:`icosahedral <orthopoly.spherical_harmonic.grid_icosahedral>`, and :func:`Fibonacci <orthopoly.spherical_harmonic.grid_fibonacci>`) and for creating random spherical harmonic expansions with specific power density relationships (noise). The module does not have functions for performing spherical harmonic analysis (transforming from values on the sphere to expansion coefficients).

For some applications, fitting a spherical harmonic expansion to data in spherical coordinates is useful. A least squares fit can be computed with the pseudoinverse of a matrix full of spherical harmonic function evaluations (see :func:`~orthopoly.spherical_harmonic.sph_har_matrix` and related functions). However, this should only be done when the number of points is much greater than the number of terms in the fitted expansion.

The books cited above have some good discussion of spherical harmonics. Other useful sources include:

* Press, William H., et al. *Numerical recipes 3rd edition: The art of scientific computing*. Cambridge university press, 2007.
* Dahlen, F., and Jeroen Tromp. *Theoretical global seismology*. Princeton university press, 1998.
* Bosch, W. "On the computation of derivatives of Legendre functions." Physics and Chemistry of the Earth, Part A: Solid Earth and Geodesy 25.9-11 (2000): 655-659.
"""

from . import util
from . import chebyshev
from . import gegenbauer
from . import legendre
from . import spherical_harmonic

__all__ = ['util', 'chebyshev', 'gegenbauer', 'legendre', 'spherical_harmonic']



"""
This module implements the Associated Legendre Polynomials, :math:`P_n^m(x)`, and their first two derivatives in support of the :mod:`~orthopoly.spherical_harmonic` module. If :math:`m=0`, they reduce to the unassociated Legendre polynomials.
"""

__all__ = [
    'legen_norm',
    'legen_hat',
    'legen',
    'legen_theta',
    'dlegen_theta',
    'ddlegen_theta',
]

from numpy import *

from .util import *

def _check_legen_hat_args(xhat, n, m):
    """Checks arguments to Legendre functions are valid"""

    #check points within bounds
    assert all(abs(xhat) <= 1.0), "can't evaluate associated Lege ndre outside [-1,1]"
    #check order isn't negative
    assert m >= 0, "can't have m < 0 in evaluating associated Legendre"
    #warn about order greater than degree
    if m > n:
        warnings.warn("A Legendre function of order greater than degree (m > n) was evaluated. These are zero by the definition and it's unusual to need them.")
        return(0.0*xhat)

def legen_norm(n, m):
    r"""Evaluates the normalization factor for the associated Legendre polynomials,

    .. math::
        :nowrap:

        \begin{equation}
            \sqrt{\frac{(2n + 1) (n - m)!}{4 \pi (n + m)!}}
        \end{equation}

    with an iteration instead of direct factorials to help avoid overflow. This function is not used in the other functions in this module, which compute the normalized polynomials without directly calculating the normalization factor. However, it can be used to switch between normalized and unnormalized values.

    :param int n: degree of polynomial
    :param int m: order of polynomial (m >= 0)

    :return: normalization factor"""

    f = sqrt((2*n + 1)/(4*pi))
    for i in range(n - m + 1, n + m + 1):
        f /= sqrt(float(i))

    return(f)

def legen_hat(xhat, n, m):
    r"""Evaluates the normalized associated Legendre function of degree n and order m, :math:`P_n^m(x)`, through a three term recurrence relationship. These are **normalized** and the normalization factor is:

    .. math::
        :nowrap:

        \begin{equation}
            \sqrt{\frac{(2n + 1) (n - m)!}{4 \pi (n + m)!}}
        \end{equation}

    as defined in chapter 6 of the reference below, but without the extra factor of :math:`-1^m`
        * Press, William H., et al. Numerical recipes 3rd edition: The art of scientific computing. Cambridge university press, 2007.

    :param array/float xhat: evaluation point in :math:`[-1,1]` (can be an array)
    :param int n: degree of polynomial
    :param int m: order of polynomial (m >= 0)

    :return: evaluated function, :math:`P_n^m(\hat{x})`"""

    #check for issues with input
    _check_legen_hat_args(xhat, n, m)

    #evaluate the initial recursion value, P_m^m, with stable factorials
    pprev = 1.0 + 0.0*xhat #running value for factorials, which becomes P_m^m
    if m > 0: #account for 0! and avoid division by zero
        pprev /= sqrt(2.0*m)
        for i in range(2*m-1, 1, -1):
            if (2*m - 1 - i) % 2 == 0: #double factorial only on every other
                pprev *= i
            pprev /= sqrt(i)
    pprev *= sqrt((2.0*m + 1.0)/(4.0*pi))*((1.0 - xhat**2)**(m/2.0))
    if n == m: return(pprev)
    #evaluate the second value
    pcurr = xhat*sqrt(2.0*m + 3.0)*pprev
    if n == m + 1: return(pcurr)
    #do the recursion
    i = m + 2
    while i <= n:
        #prefactors
        a = sqrt((4.0*i**2 - 1.0)/(i**2 - m**2))
        b = sqrt(((i - 1.0)**2 - m**2)/(4.0*(i - 1.0)**2 - 1.0))
        #evaluate and swap simultaneously
        pprev, pcurr = pcurr, a*(xhat*pcurr - b*pprev)
        #increment
        i += 1

    return(pcurr)

def legen(x, n, m, xa=-1, xb=1):
    r"""Evaluates the normalized associated Legendre function of degree n and order m, :math:`P_n^m(x)`, through a three term recurrence relationship, over the interval :math:`[x_a,x_b]`. These are **normalized** and the normalization factor is:

    .. math::
        :nowrap:

        \begin{equation}
            \sqrt{\frac{(2n + 1) (n - m)!}{4 \pi (n + m)!}}
        \end{equation}

    as defined in chapter 6 of the reference below, but without the extra factor of :math:`-1^m`
        * Press, William H., et al. Numerical recipes 3rd edition: The art of scientific computing. Cambridge university press, 2007.

    :param array/float x: evaluation point in :math:`[x_a,x_b]` (can be an array)
    :param int n: degree of polynomial
    :param int m: order of polynomial (m >= 0)
    :param float xa: lower limit of evaluation interval
    :param float xb: upper limit of evaluation interval

    :return: evaluated function, :math:`P_n^m(x)`"""

    #map the evaluation points to [-1,1]
    xhat = x2xhat(x, xa, xb)
    #evaluate
    return( legen_hat(xhat, n, m) )

def legen_theta(t, n, m):
    r"""Evaluates the normalized associated Legendre function :math:`P_n^m(cos(\theta))` with a colatitude argument in :math:`[0,\pi]` instead of :math:`[-1,1]`

    :param array/float t: colatitude evaluation point(s) in :math:`[0,\pi]` (can be an array)
    :param int n: degree of polynomial
    :param int m: order of polynomial (m >= 0)

    :return: evaluated function, :math:`P_n^m[\cos(\theta)]`"""

    #call the regular Legendre function with xhat=cos(theta)
    return( legen_hat(cos(t), n, m) )

def dlegen_theta(t, n, m):
    r"""Evaluates the first derivative of the normalized associated Legendre function with colatitude argument, :math:`d P_n^m / d \theta`. This can be used in computing the gradient of spherical harmonics. The algorithm is detailed in:
        * Bosch, W. "On the computation of derivatives of Legendre functions." Physics and Chemistry of the Earth, Part A: Solid Earth and Geodesy 25.9-11 (2000): 655-659.

    :param array/float t: colatitude evaluation point(s) in :math:`[0,\pi]` (can be an array)
    :param int n: degree of polynomial
    :param int m: order of polynomial (m >= 0)

    :return: evaluated first derivative, :math:`d P_n^m / d \theta`"""

    if m == 0 and n == 0:
        return( 0.0*t )
    elif m == 0:
        return( -sqrt(n*(n+1.0))*legen_theta(t,n,1) )
    elif m == n:
        return( sqrt(n/2)*legen_theta(t,n,n-1) )
    else:
        return( (sqrt((n+m)*(n-m+1))*legen_theta(t,n,m-1)
               - sqrt((n+m+1)*(n-m))*legen_theta(t,n,m+1))/2 )

def ddlegen_theta(t, n, m):
    r"""Evaluates the second derivative of the normalized associated Legendre function with colatitude argument, :math:`d^2 P_n^m / d \theta^2`. This can be used in computing the Laplacian of spherical harmonics. The algorithm is detailed in:
        * Bosch, W. "On the computation of derivatives of Legendre functions." Physics and Chemistry of the Earth, Part A: Solid Earth and Geodesy 25.9-11 (2000): 655-659.

    :param array/float t: colatitude evaluation point(s) in :math:`[0,\pi]` (can be an array)
    :param int n: degree of polynomial
    :param int m: order of polynomial (m >= 0)

    :return: evaluated second derivative, :math:`d^2 P_n^m / d \theta^2`"""

    if m == 0 and n == 0:
        return( 0.0*t )
    elif m == 0:
        return( -sqrt(n*(n+1))*dlegen_theta(t,n,1) )
    elif m == n:
        return( sqrt(n/2)*dlegen_theta(t,n,n-1) )
    else:
        return( (sqrt((n+m)*(n-m+1))*dlegen_theta(t,n,m-1)
               - sqrt((n+m+1)*(n-m))*dlegen_theta(t,n,m+1))/2 )



r"""
The chebyshev module is a collection of functions for setting up and using Chebyshev expansions, which can be used for very high accuracy interpolation on smooth problems and for numerical solutions to boundary value problems and PDEs. Since you're reading this, I assume you already know how awesome Chebyshev expansions are! For the right problems, they have exponential, or "infinite order," convergence!

Anyway, Chebyshev polynomials (of the first kind) can be defined and evaluated in a number of ways. Here they are defined by

.. math::   T_k(\hat{x}) = \cos(k \arccos(\hat{x})) \qquad \hat{x} \in [-1,1] \, ,

where :math:`k` is the degree of the polynomial and :math:`\hat{x}`. The polynomials are orthogonal with respect to the weighted inner product

.. math:: \left< f(x),g(x) \right> = \int_{-1}^1 f(x) g(x) \frac{1}{\sqrt{1 - x^2}} dx \, .

Through a quick transformation, the polynomials can be mapped from :math:`[-1,1]` to some other interval :math:`[x_a,x_b]` using

.. math::   x(\hat{x}) = \frac{\hat{x} + 1}{2} (x_b - x_a) + x_a \qquad x \in [x_a,x_b]

and back again by

.. math::   \hat{x}(x) = 2 \frac{x - x_a}{x_b - x_a} - 1 \qquad \hat{x} \in [-1,1].

It's often convenient to use a third coordinate, :math:`\theta`, because of the trigonometric definition of the polynomials. For :math:`\theta = \cos(\hat{x})`, the Chebyshev polynomials are

.. math::   T_k(\cos(\hat{x})) = T_k(\theta) = \cos(k\theta) ,

simply cosines, where :math:`\theta` is in the interval :math:`[\pi,0]`. Using :math:`\theta` coordinates in addition to the :math:`\hat{x}` and :math:`x` coordinates simplifies derivatives of the polynomials and enables the use of discrete cosine transformations to obtain expansion coefficients, as implemeneted in the :func:`cheby_dct` function.

The first couple derivatives of the polynomials can be evaluated with respect to :math:`\theta` by using the chain rule to convert to :math:`\hat{x}` coordinates. Derivatives then must be scaled to :math:`[x_a,x_b]` by dividing by :math:`((x_b - x_a)/2)^p`, where :math:`p` is the order of the derivative.

In addition to functions for evaluating the Chebyshev polynomials and their derivatives, some higher level functions are provided by this module.
    1. :func:`cheby_bvp_setup` creates the grid and matrix used to solve boundary value problems in 1D, capable of handing boundary conditions of arbitrary derivative order and 1st or 2nd internal derivatives.
    2. :func:`cheby_coef_setup` also creates a grid and matrix, but only to solve for the expansion coefficients, without derivatives. This is just a transform from some set of points to a set of Chebyshev polynomials. It allows for the use of a first derivative on one of the boundaries and for extra points in the grid.
    3. :func:`cheby_dct_setup` creates a grid for generating chebyshev expansions using the discrete cosine transform (DCT). It does not allow for derivatives or extra points, however. The underlying function for performing a discrete chebyshev transform is :func:`cheby_dct` and the inverting function is :func:`cheby_idct`.

Tons of information on these and other methods can be found in these references:
    * Boyd, John P. *Chebyshev and Fourier spectral methods*. Courier Corporation, 2001.
    * Fornberg, Bengt. *A practical guide to pseudospectral methods. Vol. 1*. Cambridge university press, 1998.
    * Canuto, Claudio, et al. *Spectral methods*. Springer-Verlag, Berlin, 2006.
"""

from numpy import *
from scipy.fftpack import dct, idct

from .util import *

__all__ = [
    'insert_points',
    'cheby_grid',
    'cheby',
    'cheby_hat',
    'cheby_hat_ext',
    'dcheby_t',
    'ddcheby_t',
    'dpcheby_boundary',
    'cheby_sum',
    'dcheby_t_sum',
    'cheby_hat_sum',
    'cheby_hat_ext_sum',
    'dcheby_coef',
    'cheby_hat_recur_sum',
    'dcheby_hat_recur_sum',
    'cheby_dct',
    'cheby_idct',
    'cheby_bvp_setup',
    'cheby_coef_setup',
    'cheby_dct_setup'
]

#-------------------------------------------------------------------------------
# grid functions

def insert_points(xhat, theta, x, xextra, xa, xb):
    r"""Puts some extra points into the cheby grid, ignoring duplicates and keeping things correctly sorted

    :param array xhat: grid points on :math:`[-1,1]`
    :param array theta: grid points on :math:`[\pi,0]`
    :param array x: grid points on :math:`[x_a,x_b]`
    :param array xextra: the x points to insert
    :param float xa: left boundary
    :param float xb: right boundary

    :return: tuple containing

        - grid points on :math:`[-1,1]` with points inserted
        - grid points on :math:`[\pi,0]` with points inserted
        - grid points on :math:`[x_a,x_b]` with points inserted"""

    xextra = array(xextra)
    assert all((x >= xa) | (x <= xb)), 'extra points cannot be outside the domain [xa,xb]'

    x = sort(array(list(set(x) | set(xextra))))
    xhat = x2xhat(x, xa, xb)
    theta = arccos(xhat)

    return(xhat, theta, x)

def cheby_grid(xa, xb, n):
    r"""Computes the chebyshev "extreme points" for use with all the functions in this module

    The grid points are in :math:`[-1,1]` and they are

    .. math:: \hat{x}_i = \cos\left(\frac{\pi j}{n-1}\right) \qquad j = n-1,...,0 \, .

    These points are returned with their counterparts in :math:`\theta` and :math:`x` coordinates.

    :param float xa: value of left (lower) domain boundary
    :param float xb: value of right (higher) domain boundary
    :param int n: number of points to use in the domain (including boundary points)

    :return: tuple containing

        - array of collocation points (:math:`\hat{x}` points) in :math:`[-1,1]`
        - array of theta values, :math:`\arccos(\hat{x})`, in :math:`[0,\pi]`
        - array collocation points in :math:`[x_a,x_b]`"""

    #order of expansion
    ord = n - 1
    #grid (collocation/interpolation points)
    j = arange(0, n)
    theta = (pi*j/ord)[::-1]
    xhat = cos(theta) #points in [-1,1]
    x = xhat2x(xhat, xa, xb) #points in [xa, xb]

    return(xhat, theta, x)

#-------------------------------------------------------------------------------
# evaluating Chebyshev polynomials and their derivaties

cheby = lambda x, k, xa=-1, xb=1: cos(k*arccos(x2xhat(x, xa, xb)))
cheby.__doc__ = r'Evaluates the :math:`k^{\textrm{th}}` chebyshev polynomial in :math:`[x_a,x_b]` at :math:`x`'

cheby_hat = lambda xhat, k: cos(k*arccos(xhat))
cheby_hat.__doc__ = r'Evaluates the :math:`k^{\textrm{th}}` chebyshev polynomial in :math:`[-1,1]`'

dcheby_t = lambda t, k: k*sin(k*t)/sin(t)
dcheby_t.__doc__ = r'Evaluates the first derivative of the :math:`k^{\textrm{th}}` order chebyshev polynomial at theta values'

ddcheby_t = lambda t, k: (sin(t)*(-k*k*cos(k*t)) - cos(t)*(-k*sin(k*t)))/(sin(t)**3)
ddcheby_t.__doc__ = r'Evaluates the second derivative of the :math:`k^{\textrm{th}}` order chebyshev polynomial at theta values'

def cheby_hat_ext(xhat, k):
    r"""Evaluates a chebyshev polynomial in :math:`\hat{x}` space, but potentially outside of :math:`[-1,1]`

    :param float/array xhat: argument
    :param int k: polynomial degree

    :return: :math:`T_k(\hat{x})`"""

    if abs(xhat) <= 1.0:
        return( cheby_hat(xhat, k) )
    else:
        return( ((xhat - sqrt(xhat**2 - 1))**k + (xhat + sqrt(xhat**2 - 1))**k)/2 )

def dpcheby_boundary(sign, k, p):
    r"""Evaluates the :math:`p^{\textrm{th}}` derivative of the :math:`k^{\textrm{th}}` order chebyshev polynomial at -1 or 1

    :param int sign: -1 or 1
    :param int k: polynomial degree
    :param int p: derivative order

    :return: :math:`d T_k(\pm 1) / dx`"""

    assert sign == 1 or sign == -1, 'sign must be -1 or 1'
    n = arange(0,p)
    d = (sign**(k+p))*prod((k**2 - n**2)/(2*n + 1))
    return(d)

#-------------------------------------------------------------------------------
# evaluating Chebyshev expansions and their derivaties

cheby_sum = lambda x, a, xa, xb: sum(a[k]*cheby(x, k, xa, xb) for k in range(len(a)))
cheby_sum.__doc__ = 'Evaluates a chebyshev expansion in :math:`[x_a,x_b]` with coefficient array :math:`a_i` by direct summation of the polynomials'

dcheby_t_sum = lambda t, a: sum(a[k]*dcheby_t(t, k) for k in range(len(a)))
dcheby_t_sum.__doc__ = r'Evaluates the first derivative of the cheby expansion at :math:`\theta` values'

cheby_hat_sum = lambda xhat, a: sum(a[i]*cheby_hat(xhat, i) for i in range(len(a)))
cheby_hat_sum.__doc__ = 'Evaluates a chebyshev series in :math:`[-1,1]` with coefficient array :math:`a_i` by direct summation of the polynomials'

def cheby_hat_ext_sum(xhat, a):
    r"""Evaluates chebyshev expansion anywhere, including outside :math:`[-1,1]`

    :param float/array xhat: argument
    :param array a: expansion coefficients :math:`a_i`

    :return: evaluated expansion"""

    if abs(xhat) < 1.0:
        return( cheby_hat_sum(xhat, a) )
    else:
        return( sum( (a[k]*(
                  (xhat - sqrt(xhat**2 - 1))**k
                + (xhat + sqrt(xhat**2 - 1))**k)/2)
                for k in range(len(a))
            )
        )

def dcheby_coef(a):
    """Computes the coefficents of the derivative of a Chebyshev expansion using a recurrence relation. Higher order differentiation (using this function repeatedly on the same original coefficents) is mildly ill-conditioned. See these references for more info:
        * Boyd, John P. Chebyshev and Fourier spectral methods. Courier Corporation, 2001.
        * Press, William H., et al. Numerical Recipes: The Art of Scientific Computing. 3rd ed., Cambridge University Press, 2007.

    :param array a: expansion coefficients :math:`a_i`

    :return: expansion coefficients of the input expansion's derivative"""

    n = len(a) #number of coefficients
    c = zeros((n,))
    c[n-2] = 2*(n-1)*a[n-1]
    for i in range(n-2,0,-1):
        c[i-1] = c[i+1] + 2*i*a[i]
    c[0] /= 2
    return(c[:-1])

def cheby_hat_recur_sum(xhat, a):
    r"""Evaluates a Chebyshev expansion using a recurrence relationship. This is helpful for things like root finding near the boundaries of the domain because, if the terms are evaluated with :math:`\arccos`, they are not defined outside the domain. Any tiny departure outside the boundaries triggers a NaN. This recurrence is widely cited, but of course, See
        * Boyd, John P. Chebyshev and Fourier spectral methods. Courier Corporation, 2001.

    :param float/array xhat: argument
    :param array a: expansion coefficients :math:`a_i`

    :return: evaluated expansion"""

    #first two terms
    c0, c1 = 1.0, xhat
    #apply them
    y = a[0]*c0 + a[1]*c1
    #recurr
    for i in range(2,len(a)):
        c2 = 2*xhat*c1 - c0
        y += c2*a[i]
        c0 = c1
        c1 = c2
    return(y)

def dcheby_hat_recur_sum(xhat, a):
    """Computes the derivative of a Chebyshev expansion in :math:`[-1,1]` using a recurrence relation, avoiding division by zero at the boundaries when using :func:`dcheby_t`. This is the same procedure as in :func:`dcheby_coef`, but using the derivative coefficients to evaluate the derivative as the recurrence proceeds instead of just storing and returning them.

    :param float/array xhat: argument
    :param array a: expansion coefficients :math:`a_i`

    :return: evaluated expansion"""

    n = len(a) #number of coefficients
    dy = 0.0
    c2 = 0.0 #recurrence group
    c1 = 2*(n-1)*a[n-1] #initial, highest order coefficient
    dy += c1*cheby_hat_ext(xhat, n-2)
    for i in range(n-2,1,-1):
        c0 = c2 + 2*i*a[i]
        dy += c0*cheby_hat_ext(xhat, i-1)
        c2 = c1
        c1 = c0
    dy += (c2 + 2*a[1])/2.0
    return(dy)

"""def cheby_matrix(xhat, kmax, deriv=0):

    assert 0 <= deriv <= 2

    n, m = len(xhat), kmax
    A = zeros((n, m))

    for i in range(n):
        for k in range(kmax):
            if xhat == 1 or xhat == -1:
                if deriv == 0:
                    A[i,k] ="""

#-------------------------------------------------------------------------------
# discrete Chebyshev transform

def cheby_dct(y):
    r"""Finds Chebyshev expansion coefficients :math:`a_k` using a discrete cosine transform (DCT). For high order expansions, this should be much faster than computing coefficients by solving a linear system.

    The input vector is assumed to contain the values for points lying on a :func:`chebyshev grid <cheby_grid>`. The Chebyshev expansion is assumed to have the form

    .. math:: \sum_{k=0}^{n-1} a_k \cos(k \arccos(\hat{x})) \qquad \hat{x} \in [-1,1]

    However, because the Type I DCT in `scipy.fftpack` has a slightly different convention, some small modifications are made to the results of `scipy.fftpack.dct` to achieve coefficents appropriate for the expansion above.

    :param array y: values of points on a :func:`chebyshev grid <cheby_grid>`, in the proper order

    :return: array of Chebyshev coefficients :math:`a_k`"""

    #length of transform
    n = len(y)
    #transform
    a = dct(y, type=1, n=n)/(2*n - 2)
    #a little bit of correction for conventions
    a[1:-1] *= 2
    a[1::2] *= -1

    return(a)

def cheby_idct(a):
    r"""Evaluates a chebyshev series using the inverse discrete cosine transform. For high order expansions, this should be much faster than direct summation of the polynomials.

    The input vector is assumed to contain the values for points lying on a :func:`chebyshev grid <cheby_grid>`. The Chebyshev expansion is assumed to have the form

    .. math:: \sum_{k=0}^{n-1} a_k \cos(k \arccos(\hat{x})) \qquad \hat{x} \in [-1,1]

    Because of varying conventions, this function should only be used with chebyshev expansion coefficients computed with the functions in this package (:func:`cheby_coef_setup` and :func:`cheby_dct`). The Type I IDCT from `scipy.fftpack` is used (with minor modifications).

    :param array a: chebyshev expansion coefficients :math:`a_k`

    :return: array of evaluated expansion values at :func:`chebyshev grid <cheby_grid>` points"""

    #copy and correct for convention
    a = a.copy()
    a[1:-1] /= 2
    a[1::2] *= -1
    #transform
    y = idct(a, type=1)

    return(y)

#-------------------------------------------------------------------------------
# high-level functions for setting up transforms

def cheby_bvp_setup(xa, xb, n, aderiv, ideriv, bderiv, xextra=None):
    r"""Set up the grid and matrices for a Chebyshev spectral solver for boundary value problems in 1 cartesian dimension

    :param float xa: value of left (lower) domain boundary
    :param float xb: value of right (higher) domain boundary
    :param int n: number of points to use in the domain (including boundaries)
    :param int aderiv: order of derivative for left boundary condition (can be 0)
    :param int ideriv: order of derivative for internal nodes (must be 1 or 2)
    :param int bderiv: order of derivative for right boundary condition (can be 0)
    :param xextra: extra points within the domain to include as collocation points

    :return: tuple containing

        - collocation points in :math:`[-1,1]`
        - theta values, :math:`\arccos(\hat{x})`, in :math:`[0,\pi]`
        - collocation points in :math:`[x_a,x_b]`
        - matrix to solve for the expansion coefficients"""

    assert aderiv == 0 or bderiv == 0, 'either aderiv or bderiv must be zero'
    assert (ideriv > 0 and ideriv < 3), "ideriv must be 1 or 2"

    #scale factor for derivatives when mapping [-1,1] to [xa, xb]
    scale = (xb - xa)/2.0 #length of [xa,xb] divided by length of [-1,1]

    #grid points
    xhat, theta, x = cheby_grid(xa, xb, n)
    if xextra is not None:
        xhat, theta, x = insert_points(xhat, theta, x, xextra, xa, xb)
        n = len(x)

    #solver matrix (j is for rows/points and k is for colums or expansion order)
    A = zeros((n,n))
    #left boundary condition/row
    if aderiv > 0:
        for k in range(n): A[0,k] = dpcheby_boundary(-1, k, aderiv)/(scale**aderiv)
    else:
        for k in range(n): A[0,k] = cheby_hat(-1, k)
    #internal nodes
    if ideriv == 1:
        dfunk = dcheby_t
    elif ideriv == 2:
        dfunk = ddcheby_t
    for j in range(1,n-1):
        for k in range(n):
            #have to do the js backward if the x values are to be in increasing order
            A[j,k] = dfunk(theta[j], k)/(scale**ideriv)
    #right boundary condition/row
    if bderiv > 0:
        for k in range(n): A[-1,k] = dpcheby_boundary(1, k, bderiv)/(scale**bderiv)
    else:
        for k in range(n): A[-1,k] = cheby_hat(1, k)

    return(xhat, theta, x, A)

def cheby_coef_setup(xa, xb, n, da=False, db=False, xextra=None):
    r"""Constructs the grid and matrix for finding the coefficients of the chebyshev expansion on :math:`[x_a,x_b]` with :math:`n` points

    :param float xa: value of left (lower) domain boundary
    :param float xb: value of right (higher) domain boundary
    :param int n: number of points to use in the domain (including boundaries)
    :param bool da: use the first deriv at the left/low boundary as the boundary condition
    :param bool da: use the first deriv at the right/high boundary as the boundary condition
    :param array xextra: extra points within the domain to include as collocation points

    :return: tuple containing

        - collocation points in :math:`[-1,1]`
        - theta values, :math:`\arccos(\hat{x})`, in :math:`[0,\pi]`
        - collocation points in :math:`[x_a,x_b]`
        - matrix to solve for the expansion coefficients
        - scale factor for derivatives"""

    #scale factor for derivatives when mapping [-1,1] to [xa, xb]
    scale = (xb - xa)/2.0 #length of [xa,xb] divided by length of [-1,1]

    #grid points
    xhat, theta, x = cheby_grid(xa, xb, n)
    if xextra is not None:
        xhat, theta, x = insert_points(xhat, theta, x, xextra, xa, xb)
        n = len(x)

    #solver matrix (j is for rows/points and k is for colums or expansion order)
    A = zeros((n,n))
    for j in range(n):
        for k in range(n):
            A[j,k] = cheby_hat(xhat[j], k)
    if da:
        for k in range(n): A[0,k] = dpcheby_boundary(-1, k, 1)/scale
    if db:
        for k in range(n): A[-1,k] = dpcheby_boundary(1, k, 1)/scale

    return(xhat, theta, x, A, scale)

def cheby_dct_setup(xa, xb, n):
    r"""Constructs a Chebyshev grid and return a function for computing Chebyshev expansion coefficients on the grid with a discrete cosine transform (DCT). The returned function wraps :func:`cheby_dct`.

    :param float xa: value of left (lower) domain boundary
    :param float xb: value of right (higher) domain boundary
    :param int n: number of points to use in the domain (including boundaries)

    :return: tuple containing

        - collocation points in :math:`[-1,1]`
        - theta values, :math:`\arccos(\hat{x})`, in :math:`[0,\pi]`
        - collocation points in :math:`[x_a,x_b]`
        - function for computing expansion coefficients
        - scale factor for derivatives"""

    #scale factor for derivatives when mapping [-1,1] to [xa, xb]
    scale = (xb - xa)/2.0 #length of [xa,xb] divided by length of [-1,1]

    #grid points
    xhat, theta, x = cheby_grid(xa, xb, n)

    #DCT function for computing coefficents
    func = lambda y: cheby_dct(y)

    return(xhat, theta, x, func, scale)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="orthopoly",
    version="0.9.3",
    author="Mark Baum <markmbaum@protonmail.com>",
    author_email="markmbaum@protonmail.com",
    description="Python functions for orthogonal polynomials and (real, 2D, orthonormal) spherical harmonics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/markmbaum/orthopoly",
    packages=setuptools.find_packages(),
    python_requires='>=3',
    install_requires=[
        'numpy',
        'scipy>=1.5.4'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
General use functions
"""

from numpy import *

__all__ = [
    'isnum',
    'xhat2x',
    'x2xhat',
    'xhat2theta',
    'theta2xhat',
]

def isnum(x):
    """Tests if an object is float-able (a number)

    :param x: some object

    :return: True or False"""

    try:
        x = float(x)
    except ValueError:
        return(False)
    else:
        return(True)

#-------------------------------------------------------------------------------
# coordinate mappings/conversions

xhat2x = lambda xhat, a, b: (xhat + 1.0)*((b - a)/2.0) + a
xhat2x.__doc__ = r'Converts :math:`\hat{x}` coordinates in :math:`[-1,1]` to :math:`x` coordinates in :math:`[x_a,x_b]`'

x2xhat = lambda x, a, b: 2.0*(x - a)/(b - a) - 1.0
x2xhat.__doc__ = r'Converts :math:`x` coordinates in :math:`[x_a,x_b]` to :math:`\hat{x}` coordinates in :math:`[-1,1]`'

xhat2theta = lambda xhat: arccos(xhat)
xhat2theta.__doc__ = r'Converts :math:`\hat{x}` coordinates in :math:`[-1,1]` to :math:`\theta` coordinates :math:`[\pi,0]`'

theta2xhat = lambda xhat: arccos(xhat)
theta2xhat.__doc__ = r'Converts :math:`\theta` coordinates in :math:`[\pi,0]` to :math:`\hat{x}` coordinates in :math:`[-1,1]`'
